program lookup;

{
 lookup - nachschlagen mit Index
 Copyright (c) 12/01 by Sebastian Koppehel, <basti@bastisoft.de>
 }

uses
   getopts;

const
   DEF_INDEX_NAME = 'index.dat';
   DEF_DICT_NAME  = 'dict.txt';
   FILE_BUF	  = 255;
{$IFDEF FPC}
   VERSION  = 'lookup 1 ' + {$I %DATE%};
{$ELSE}
   VERSION = 'lookup 1';
{$ENDIF}
   LEV_FULL = 0;
   LEV_KEYS = 1;
   LEV_LINES = 2;

const
   L_OPTS : array[1..7] of TOption = 
   ((name : 'index'; has_arg : required_argument; flag : nil; value : 'i'),
    (name : 'dict'; has_arg : required_argument; flag : nil; value : 'd'),
    (name : 'case'; has_arg : no_argument; flag : nil; value : 'c'),
    (name : 'version'; has_arg : no_argument; flag : nil; value : 'v'),
    (name : 'keys'; has_arg : no_argument; flag : nil; value : 'k'),
    (name : 'lines'; has_arg : no_argument; flag : nil; value : 'l'),
    (name : ''; has_arg : 0; flag : nil; value : #0));
   S_OPTS = 'i:d:cklv';

type
   pposr = ^posr;
   posr	 = record
	      pos  : longint;
	      next : pposr;
	   end;	   
   node	 = record
	      name : string;
	      pos  : pposr;
	      bit  : longint;
	      l, r : longint;
	   end;	   
   carr	= array[1..MAXLONGINT] of byte;

var
   fin	   : file;
   maxlen  : longint;
   maxpos  : longint;
   buf	   : ^carr;
   bufsize : longint;

procedure read_node(var nd : node; index : longint);
var nlen : longint;
begin
   //writeln(index);
   seek(fin, 2 * sizeof(longint) + index);
   blockread(fin, buf^, bufsize);
   with nd do
   begin
      nlen := ord(buf^[1]);
      setlength(name, nlen);
      move(buf^[2], name[1], length(name));
      move(buf^[nlen + 2], bit, sizeof(longint));
      move(buf^[nlen + 2 + sizeof(longint)], l, sizeof(longint));
      move(buf^[nlen + 2 + 2 * sizeof(longint)], r, sizeof(longint));
   end;
end;

procedure getps(var nd : node);
var
   numpos, i, nlen : longint;
   pr		   : pposr;
begin
   nlen := length(nd.name) + 1;
   move(buf^[nlen + 1 + 3 * sizeof(longint)], numpos, sizeof(longint));
   new(nd.pos);
   pr := nd.pos;
   move(buf^[nlen + 1 + 4 * sizeof(longint)], pr^.pos, sizeof(longint));
   pr^.next := nil;

   for i := 1 to numpos - 1 do
   begin
      new(pr^.next);
      pr := pr^.next;
      move(buf^[nlen + 1 + (4 + i) * sizeof(longint)],
	   pr^.pos, sizeof(longint));
      pr^.next := nil;
   end;
end;

function getbit(s : string; bit : longint) : longint;
var c : longint;
begin
   c := bit div 8;
   if c <= length(s) then getbit := ord(s[c]) shr (bit mod 8) and 1
   else getbit := 0;
end;

procedure strdown(var s : string);
const
   SMALL = '';
   CAP	 = '';
var i, p : longint;
begin
   for i := 1 to length(s) do
      if s[i] in ['A'..'Z'] then
         s[i] := chr(ord(s[i]) or $20)
      else begin
         p := pos(s[i], CAP);
         if p > 0 then s[i] := SMALL[p];
      end;
end;

procedure print_entry(var f : file; ps, level : longint; search : string);
var
   done	: boolean;
   have	: longint;
   line	: string;
   i, j	: longint;
   fbuf	: string;
   cmpl	: string;
begin
   seek(f, ps);
   j := 0;
   line := '';
   done := false;
   repeat
      setlength(fbuf, FILE_BUF);
      blockread(f, fbuf[1], FILE_BUF, have);
      setlength(fbuf, have);
      repeat
	 i := 0;
	 while (i < length(fbuf)) and not (fbuf[i + 1] in [#10,#13]) do inc(i);
	 line := line + copy(fbuf, 1, i);
	 if i < length(fbuf) then
	 begin
	    inc(j);
	    if level = LEV_KEYS then
	    begin
	       writeln(line);
	       done := true;
	    end
	    else if j > 1 then
	    begin
	       if line = '%' then done := true
	       else if level = LEV_LINES then
	       begin
		  cmpl := line;
		  strdown(cmpl);
		  if pos(search, cmpl) > 0 then writeln(line);
	       end
	       else writeln(line);
	    end
	    else if level = LEV_LINES then writeln(line);
	    line := '';
	 end;
	 if (i + 1 < length(fbuf)) and (fbuf[i + 1] = #13) and
	    (fbuf[i + 2] = #10) then inc(i);
	 delete(fbuf, 1, i + 1);
      until done or (length(fbuf) = 0);
   until eof(f) or done;
end;

var
   p, nd  : node;
   search : string;
   next	  : longint;
   fcorp  : file;
   i	  : longint;
   pr	  : pposr;
   sens	  : boolean;
   inname : string;
   dname  : string;
   optc	  : longint;
   c	  : char;
   level  : longint;

begin
   inname := DEF_INDEX_NAME;
   dname := DEF_DICT_NAME;
   sens := false;
   level := LEV_FULL;

   repeat
      c := GetLongOpts(S_OPTS, @L_OPTS, optc);
      case c of
	'i' : inname := optarg;
	'd' : dname := optarg;
	'c' : sens := true;
	'k' : level := LEV_KEYS;
	'l' : level := LEV_LINES;
	'v' :
              begin
		 writeln(VERSION);
		 halt;
	      end;
      end;
   until c = EndOfOptions;

   assign(fin, inname);
   reset(fin, 1);
   blockread(fin, maxlen, sizeof(maxlen));
   blockread(fin, maxpos, sizeof(maxpos));
   bufsize := maxlen + 1 + 3 * sizeof(longint) +
              (maxpos + 1) * sizeof(longint);
   getmem(buf, bufsize);

   assign(fcorp, dname);
   reset(fcorp, 1);

   for optc := optind to paramcount do
   begin
      search := paramstr(optc);
      if not sens then strdown(search);

      next := 0;
      read_node(nd, next);
      repeat
	 p := nd;
	 if getbit(search, nd.bit) = 0 then next := nd.l else next := nd.r;
	 read_node(nd, next);
      until p.bit <= nd.bit;

      if nd.name = search then
      begin
	 getps(nd);
	 pr := nd.pos;
	 while pr <> nil do
	 begin
	    print_entry(fcorp, pr^.pos, level, search);
	    // writeln(pr^.pos);
	    pr := pr^.next;
	    if (pr <> nil) and (level <> LEV_KEYS) then
	    begin
	       for i := 1 to 39 do write('- ');
	       writeln;
	    end;
	 end;
      end
      else writeln('"', search, '" nicht gefunden.');
      if optc < paramcount then
      begin
	 for i := 1 to 78 do write('-');
	 writeln;
      end;
   end;
   close(fin);
   close(fcorp);
end.
