program mki;

uses
   getopts;

type
   pposr = ^posr;
   posr	 = record
	      pos  : longint;
	      next : pposr;
	   end;	   
   pnode = ^node;  
   node	 = record
	      name : string;
	      pos  : pposr;
	      bit  : longint;
	      l, r : pnode;
	      find : longint;
	   end;	   
   pwait = ^wait;
   wait	 = record
	      nd   : pnode;
	      next : pwait;
	   end;	   

const 
   DEF_MAXBIT	  = 800;
   DEF_INDEX_NAME = 'index.dat';
   DEF_DICT_NAME  = 'dict.txt';
{$IFDEF FPC}
   VERSION	  = 'mki 1, ' + {$I %DATE%};
{$ELSE}
   VERSON         = 'mki 1';
{$ENDIF}

const
   L_OPTS : array[1..8] of TOption = 
   ((name : 'index'; has_arg : required_argument; flag : nil; value : 'i'),
    (name : 'dict'; has_arg : required_argument; flag : nil; value : 'd'),
    (name : 'full'; has_arg : no_argument; flag : nil; value : 'f'),
    (name : 'exclude'; has_arg : required_argument; flag : nil; value : 'e'),
    (name : 'case'; has_arg : no_argument; flag : nil; value : 'c'),
    (name : 'dos-format'; has_arg : no_argument; flag : nil; value : 'D'),
    (name : 'version'; has_arg : no_argument; flag : nil; value : 'v'),
    (name : ''; has_arg : 0; flag : nil; value : #0));
   S_OPTS = 'i:d:fe:cDv';

var
   maxbit : longint;
   maxpos : longint;
   maxlen : longint;

var
   maxfreq : pnode;
   longest : pnode;

procedure add_pos(nd : pnode; pos : longint);
var
   pr : pposr;
   i  : longint;
begin
   pr := nd^.pos;
   i := 0;
   while (pr^.pos <> pos) and (pr^.next <> nil) do
   begin
      pr := pr^.next;
      inc(i)
   end;

   if pr^.pos <> pos then
   begin
      new(pr^.next);
      pr^.next^.pos := pos;
      pr^.next^.next := nil;
   end;

   if i + 1 > maxpos then
   begin
      maxpos := i + 1;
      maxfreq := nd;
   end;
end;

function getbit(s : string; bit : longint) : longint;
var c : longint;
begin
   c := bit div 8;
   if c <= length(s) then getbit := ord(s[c]) shr (bit mod 8) and 1
   else getbit := 0;
end;

function find(name : string; n : pnode) : pnode;
var
   p : pnode;
begin
   repeat
      p := n;
      if getbit(name, n^.bit) = 0 then n := n^.l else n := n^.r;
   until p^.bit <= n^.bit;
   find := n;
end;

function add_node(name : string; pos : longint; n : pnode) : pnode;
var
   bit	: longint;
   p, f	: pnode;
begin
   f := find(name, n);
   if f^.name = name then
   begin
      add_pos(f, pos);
      add_node := f;
      exit;
   end;

   bit := maxbit;
   while (getbit(name, bit) = getbit(f^.name, bit)) do dec(bit);
   repeat
      p := n;
      if getbit(name, n^.bit) = 0 then n := n^.l else n := n^.r;
   until (n^.bit <= bit) or (p^.bit <= n^.bit);

   new(f);
   f^.name := name;
   new(f^.pos); f^.pos^.pos := pos; f^.pos^.next := nil;
   f^.bit := bit;

   if getbit(name, bit) = 0 then
   begin
      f^.l := f;
      f^.r := n;
   end
   else begin
      f^.l := n;
      f^.r := f;
   end;
   if getbit(name, p^.bit) = 0 then p^. l := f else p^.r := f;

   if length(name) > maxlen then
   begin
      maxlen := length(name);
      longest := f;
   end;

   add_node := f;
end;

procedure strdown(var s : string);
const
   SMALL = '';
   CAP	 = '';
var i, p : longint;
begin
   for i := 1 to length(s) do
      if s[i] in ['A'..'Z'] then
         s[i] := chr(ord(s[i]) or $20)
      else begin
         p := pos(s[i], CAP);
         if p > 0 then s[i] := SMALL[p];
      end;
end;

function get_key(var line, stw : string; full, sens : boolean) : boolean;
const alphanum = ['a'..'z', '', '', '', 'A'..'Z', '', '', '',
		  '', '0'..'9', '-'];
var i : longint;
begin
   if length(line) = 0 then
   begin
      get_key := false;
      exit;
   end;

   if full then
   begin
      i := 0;
      while (i < length(line)) and (line[i + 1] in alphanum) do inc(i);
      stw := copy(line, 1, i);
      if (length(stw) > 0) and
	 ((stw[1] = '-') or (stw[length(stw)] = '-')) then
	 stw := '';
      while (i < length(line)) and not (line[i + 1] in alphanum) do inc(i);
      delete(line, 1, i);
   end
   else begin
      i := pos(',', line);
      if i = 0 then i := length(line) + 1;
      stw := copy(line, 1, i - 1);
      delete(line, 1, i);
   end;

   if not sens then strdown(stw);
   get_key := length(stw) > 0;
end;

procedure enqueue(var last : pwait; nd : pnode);
begin
   new(last^.next);
   last^.next^.nd := nd;
   last^.next^.next := nil;
   last := last^.next;
end;

type
   tbuf	= array[1..MAXLONGINT] of char;

var
   bufsize : longint;
   buf	   : ^tbuf;

procedure write_node(var f : file; nd : pnode; var wt : pwait);
var
   i, nlen : longint;
   pr	   : pposr;
begin
   nlen := length(nd^.name) + 1;
   move(nd^.name, buf^, nlen);
   move(nd^.bit, buf^[nlen + 1], sizeof(longint));

   if nd^.l^.bit < nd^.bit then enqueue(wt, nd^.l);
   move(nd^.l^.find, buf^[nlen + 1 + sizeof(longint)], sizeof(longint));

   if nd^.r^.bit < nd^.bit then enqueue(wt, nd^.r);
   move(nd^.r^.find, buf^[nlen + 1 + 2 * sizeof(longint)], sizeof(longint));

   i := 1;
   pr := nd^.pos;
   while pr <> nil do
   begin
      inc(i);
      move(pr^.pos, buf^[nlen + 1 + (2 + i) * sizeof(longint)],
	   sizeof(longint));
      pr := pr^.next;
   end;
   dec(i);
   move(i, buf^[nlen + 1 + 3 * sizeof(longint)], sizeof(longint));

   blockwrite(f, buf^, nlen + (i + 4) * sizeof(longint));
end;

var
   root	   : pnode;
   line	   : string;
   fin	   : text;
   fout	   : file;
   wt, wtt : pwait;
   curr	   : pwait;
   attn	   : boolean;
   key	   : string;
   cnt	   : longint;
   last	   : longint;
   fbuf	   : array[1..10240] of char;
   crlen   : longint;
   fpos	   : longint;
   full	   : boolean;
   ln2	   : string;
   excl	   : pnode;
   lnr	   : longint;
   inname  : string;
   outname : string;
   excname : string;
   c	   : char;
   sens	   : boolean;
   pr	   : pposr;
   index   : longint;

begin
   maxbit := DEF_MAXBIT;
   crlen := 1;
   full := false;
   inname := DEF_DICT_NAME;
   outname := DEF_INDEX_NAME;
   excname := '';
   sens := false;

   repeat
      c := GetLongOpts(S_OPTS, @L_OPTS, cnt);
      case c of
	'i' : outname := optarg;
	'd' : inname := optarg;
	'f' : full := true;
	'e' : excname := optarg;
	'c' : sens := true;
	'D' : crlen := 2;
	'v' :
              begin
		 writeln(VERSION);
		 halt;
	      end;
      end;
   until c = EndOfOptions;

   if full and (excname <> '') then
   begin
      new(excl);
      with excl^ do
      begin
	 name := '';
	 pos := nil;
	 bit := maxbit;
	 l := excl;
	 r := excl;
      end;
      assign(fin, excname);
      reset(fin);
      while not eof(fin) do
      begin
	 readln(fin, line);
	 strdown(line);
	 add_node(line, 0, excl);
      end;
      close(fin);
   end;

   new(root);
   with root^ do
   begin
      name := '';
      pos := 0;
      bit := maxbit;
      l := root;
      r := root;
   end;

   maxpos := 1;
   maxlen := 0;
   longest := root;
   maxfreq := root;
   
   assign(fin, inname);
   settextbuf(fin, fbuf, 10240);
   reset(fin);
   cnt := 0;
   fpos := 0;
   attn := false;
   last := 0;
   lnr := 0;
   writeln('. = 1000 Zeilen');
   while not eof(fin) do
   begin
      readln(fin, line);
      fpos := fpos + length(line) + crlen;
      if full then
      begin
	 ln2 := line;
	 while get_key(ln2, key, full, sens) do
	    if find(key, excl)^.name <> key then
	    begin
	       add_node(key, last, root);
	       // writeln(key);
	       inc(cnt);
	    end;
      end
      else if attn then
      begin
	 ln2 := line;
	 attn := false;
	 while get_key(ln2, key, full, sens) do
	 begin
	    add_node(key, last, root);
	    // writeln(key);
	    inc(cnt);
	 end;
      end;
      if line = '%' then
      begin
	 attn := true;
	 last := fpos;
      end else attn := false;
      inc(lnr);
      if lnr mod 1000 = 0 then write('.');
   end;
   close(fin);
   writeln;
   writeln('Geamtzahl: ', cnt);
   writeln('Lngstes Wort: ', longest^.name, ' (', maxlen, ' Zeichen)');
   writeln('Hufigster Eintrag: "', maxfreq^.name, '" (', maxpos, 'x)');

   bufsize := maxlen + 1 + (maxpos + 4) * sizeof(longint);
   getmem(buf, bufsize);

   new(wt);
   wt^.nd := root;
   wt^.next := nil;
   curr := wt;
   index := 0;
   repeat
      curr^.nd^.find := index;
      pr := curr^.nd^.pos;
      while pr <> nil do
      begin
	 index := index + sizeof(longint);
	 pr := pr^.next;
      end;
      index := index + length(curr^.nd^.name) + 1 + sizeof(longint) * 4;

      if curr^.nd^.l^.bit < curr^.nd^.bit then enqueue(wt, curr^.nd^.l);
      if curr^.nd^.r^.bit < curr^.nd^.bit then enqueue(wt, curr^.nd^.r);
      wtt := curr;
      curr := curr^.next;
      dispose(wtt);
   until curr = nil;

   assign(fout, outname);
   rewrite(fout, 1);
   blockwrite(fout, maxlen, sizeof(longint));
   blockwrite(fout, maxpos, sizeof(longint));
   new(wt);
   wt^.nd := root;
   wt^.next := nil;
   curr := wt;
   repeat
      write_node(fout, curr^.nd, wt);
      curr := curr^.next;
   until curr = nil;
   close(fout);
end.
